/// Exact copy from the stdext crate v0.3.1
/// <https://docs.rs/stdext/0.3.1/stdext/macro.function_name.html>
#[macro_export]
macro_rules! function_name {
    () => {{
        // Okay, this is ugly, I get it. However, this is the best we can get on a stable rust.
        fn f() {}
        fn type_name_of<T>(_: T) -> &'static str {
            std::any::type_name::<T>()
        }
        let name = type_name_of(f);
        // `3` is the length of the `::f`.
        &name[..name.len() - 3]
    }};
}

/// Only get the function name itself, not the full qualifier
#[macro_export]
macro_rules! function_name_short {
    () => {{
        let name = function_name!();
        name.split("::").last().unwrap()
    }};
}

/// Like [`println!`] but also prints the current function name (including module path) and line.
///
/// # Examples
///
/// ```
/// use rustics_dbg::DBG;
/// let x = 42;
/// let y = "hello debug";
/// DBG!("{} {}", x, y);
/// ```
///
/// ```compile_fail
/// use rustics_dbg::DBG;
/// let x = 42;
/// let y = "hello debug";
/// // Since internally the macro uses concat!, you cannot inline variables into the format string.
/// DBG!("{x} {y}"); // this won't compile
/// ```
#[macro_export]
macro_rules! DBG {
    ($fmt:literal $(,$arg:expr)*) => {
        println!(concat!("{}():{} ", $fmt), $crate::function_name!(), line!() $(,$arg)*);
    };
}

#[cfg(test)]
mod tests_function_name_full_path {

    use super::*;

    trait MyTrait {
        fn trait_fn_1(&self) -> &'static str {
            function_name!()
        }

        fn trait_fn_2(&self) -> &'static str {
            function_name!()
        }
    }

    struct MyStruct;

    impl MyStruct {
        fn method(&self) -> &'static str {
            function_name!()
        }

        fn associated_fn() -> &'static str {
            function_name!()
        }
    }

    impl MyTrait for MyStruct {
        fn trait_fn_1(&self) -> &'static str {
            function_name!()
        }
    }

    #[test]
    fn full_names() {
        let h = MyStruct;
        assert_eq!(
            "<rustics_dbg::tests_function_name_full_path::MyStruct as rustics_dbg::tests_function_name_full_path::MyTrait>::trait_fn_1",
            h.trait_fn_1()
        );
        assert_eq!(
            "rustics_dbg::tests_function_name_full_path::MyTrait::trait_fn_2",
            h.trait_fn_2()
        );
        assert_eq!(
            "rustics_dbg::tests_function_name_full_path::MyStruct::method",
            h.method()
        );
        assert_eq!(
            "rustics_dbg::tests_function_name_full_path::MyStruct::associated_fn",
            MyStruct::associated_fn()
        );
    }
}

#[cfg(test)]
mod tests_function_name_short {

    use super::*;

    trait MyTrait {
        fn trait_fn_1(&self) -> &'static str {
            function_name_short!()
        }

        fn trait_fn_2(&self) -> &'static str {
            function_name_short!()
        }
    }

    struct MyStruct;

    impl MyStruct {
        fn method(&self) -> &'static str {
            function_name_short!()
        }

        fn associated_fn() -> &'static str {
            function_name_short!()
        }
    }

    impl MyTrait for MyStruct {
        fn trait_fn_1(&self) -> &'static str {
            function_name_short!()
        }
    }

    #[test]
    fn full_names() {
        let h = MyStruct;
        assert_eq!("trait_fn_1", h.trait_fn_1());
        assert_eq!("trait_fn_2", h.trait_fn_2());
        assert_eq!("method", h.method());
        assert_eq!("associated_fn", MyStruct::associated_fn());
    }
}
